--
-- plxdssm.vhd - Local bus direct slave state machine
--
-- Modules defined:
--
--     plxdssm           Local bus direct slave state machine
--
-- This state machine is designed to respond to direct slave transfers or
-- programmed DMA transfers, NOT demand mode DMA transfers.
--
-- One cycle of address decoding is permitted.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

entity plxdssm is
    port(
        clk:        in    std_logic;  -- Local bus clock
        rst:        in    std_logic;  -- Asynchronous reset, should reduce to FPGA global reset
        sr:         in    std_logic;  -- Synchronous reset (should never be asserted mid-burst)
        qlads:      in    std_logic;  -- LADS# signal qualified by address, FHOLDA and/or other decoding
        lblast:     in    std_logic;  -- Local bus LBLAST# signal, active high
        lwrite:     in    std_logic;  -- Local bus LWRITE signal, active high
        ld_oe:      out   std_logic;  -- Output enable for local bus data pins
        lready:     out   std_logic;  -- To local bus LREADY# signal, active high
        lready_oe:  out   std_logic;  -- Enable for local bus LREADY# signal
        lbterm:     out   std_logic;  -- To local bus LBTERM# signal, active high
        lbterm_oe:  out   std_logic;  -- Enable for local bus LBTERM# signal, active high
        transfer:   out   std_logic;  -- Data transfer on this clock edge
        decode:     out   std_logic;  -- Application should do further address decoding
        write:      out   std_logic;  -- 'lwrite' latched on 'qlads'
        ready:      in    std_logic;  -- Apply pulse when ready
        stop:       in    std_logic); -- Pulse to terminate burst
end plxdssm;

architecture rtl of plxdssm is
    
    type state_type is (s_idle, s_decode, s_wait, s_xfer);
    signal state, n_state: state_type;
    signal i_transfer, n_transfer: std_logic;
    signal i_lready, n_lready : std_logic;
    signal i_lready_oe, n_lready_oe : std_logic;
    signal i_lbterm, n_lbterm : std_logic;
    signal i_decode, n_decode : std_logic;
    signal i_ld_oe, n_ld_oe : std_logic;
    signal i_write, n_write : std_logic;
    signal stopping, n_stopping : std_logic;
    
begin

    transfer   <= i_transfer;
    lready     <= i_lready;
    lready_oe  <= i_lready_oe;
    lbterm     <= i_lbterm;
    lbterm_oe  <= i_lready_oe;
    decode     <= i_decode;
    ld_oe      <= i_ld_oe;
    write      <= i_write;

    state_transition : process(clk, rst)
    begin
        if rst = '1' then
            state        <= s_idle;
            i_transfer   <= '0';
            i_lready     <= '0';
            i_lready_oe  <= '0';
            i_lbterm     <= '0';
            i_decode     <= '0';
            i_ld_oe      <= '0';
            i_write      <= '0';
            stopping     <= '0';
        elsif clk'event and clk = '1' then
            if sr = '1' then
                state        <= s_idle;
                i_transfer   <= '0';
                i_lready     <= '0';
                i_lready_oe  <= '0';
                i_lbterm     <= '0';
                i_decode     <= '0';
                i_ld_oe      <= '0';
                i_write      <= '0';
                stopping     <= '0';
            else
                state        <= n_state;
                i_transfer   <= n_transfer;
                i_lready     <= n_lready;
                i_lready_oe  <= n_lready_oe;
                i_lbterm     <= n_lbterm;
                i_decode     <= n_decode;
                i_ld_oe      <= n_ld_oe;
                i_write      <= n_write;
                stopping     <= n_stopping;
            end if;
        end if;
    end process state_transition;
	
    next_state : process(
        state, 
        lblast, 
        qlads, 
        i_lbterm, 
        lwrite, 
        ready,
        stop,
        stopping,
        i_write)
    begin
        case state is
            when s_idle =>
                n_transfer <= '0';
                n_lready  <= '0';
                n_lready_oe <= '0';
                n_lbterm <= '0';
                n_ld_oe <= '0';
                n_stopping <= '0';
                if qlads = '1' then
                    n_state <= s_decode;
                    n_decode <= '1';
                    n_write <= lwrite;
                else
                    n_state <= s_idle;
                    n_decode <= '0';
                    n_write <= i_write;
                end if;
                
            when s_decode =>
                n_decode <= '0';
                n_ld_oe <= not i_write;
                n_lready_oe <= '1';
                n_write <= i_write;
                if ready = '1' then
                    n_state <= s_xfer;
                    n_lready <= '1';
                    n_lbterm <= stop;
                    n_transfer <= '1';
                    n_stopping <= '0';
                else
                    n_state <= s_wait;
                    n_lready <= '0';
                    n_lbterm <= '0';
                    n_transfer <= '0';
                    n_stopping <= stop;
                end if;

            when s_wait =>
                n_decode <= '0';
                n_ld_oe <= not i_write;
                n_lready_oe <= '1';
                n_write <= i_write;
                if ready = '1' then
                    n_state <= s_xfer;
                    n_lready <= '1';
                    n_lbterm <= stop or stopping;
                    n_transfer <= '1';
                    n_stopping <= '0';
                else
                    n_state <= s_wait;
                    n_lready <= '0';
                    n_lbterm <= '0';
                    n_transfer <= '0';
                    n_stopping <= stop or stopping;
                end if;
                
            when s_xfer =>
                n_decode <= '0';
                n_lready_oe <= '1';
                n_stopping <= '0';
                n_write <= i_write;
                if lblast = '1' or i_lbterm = '1' then
                    n_state <= s_idle;
                    n_lready <= '0';
                    n_lbterm <= '0';
                    n_transfer <= '0';
                    n_ld_oe <= '0';
                else
                    n_state <= s_xfer;
                    n_lready <= '1';
                    n_lbterm <= stop;
                    n_transfer <= '1';
                    n_ld_oe <= not i_write;
                end if;
                
        end case;
    end process next_state;

end rtl;
