--
-- zbt64_main.vhd - Local bus interface giving host access to ZBT SSRAM
--                  on an ADM-XRC series card.
--
-- (C) Copyright Alpha Data Parallel Systems Ltd. 1999-2004
--
-- This module is a parameterisable 64-bit local bus interface to ZBT
-- SSRAM. The parameters are as follows:
--
--   num_bank     - number of logical banks of SSRAM
--   num_clock    - number of SSRAM clocks to output
--   addr_width   - (maximum) number of address bits to the SSRAM
--   ctl_width    - number of control bits to the SSRAM
--   data_width   - number of data bits per physical SSRAM bank
--   bank_group   - number of physical SSRAM banks to combine in
--                  order to produce a logical SSRAM bank
--
-- When the physical SSRAM banks on the card are less than 64 bits
-- wide, two or more banks (specified by 'bank_group') may be combined
-- in order to produce a single logical bank. Thus, the product of
-- 'bank_group' and 'data_width' must equal 64.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.localbus.all;

entity zbt64_main is
    generic(
        num_bank       : in    natural;
        num_clock      : in    natural;
        addr_width     : in    natural;
        ctl_width      : in    natural;
        data_width     : in    natural;
        bank_group     : in    natural);
    port(
        lclk           : in    std_logic;
        mclk           : in    std_logic;
        ramclko        : out   std_logic_vector(num_clock - 1 downto 0);
        ramclki        : in    std_logic_vector(num_clock - 1 downto 0);
        lreset_l       : in    std_logic; 
        lwrite         : in    std_logic;
        lads_l         : in    std_logic;
        lblast_l       : in    std_logic;
        lbterm_l       : out   std_logic;
        la_i           : in    std_logic_vector(23 downto 3);
        ld_i           : in    std_logic_vector(63 downto 0);
        ld_o           : out   std_logic_vector(63 downto 0);
        oe_ld          : out   std_logic;
        lready_l       : out   std_logic; 
        lbe_l          : in    std_logic_vector(7 downto 0);
        fholda         : in    std_logic;
        ra0            : out   std_logic_vector(addr_width * bank_group - 1 downto 0);
        rc0            : out   std_logic_vector(ctl_width  * bank_group - 1 downto 0);
        rd0            : inout std_logic_vector(data_width * bank_group - 1 downto 0);
        ra1            : out   std_logic_vector(addr_width * bank_group - 1 downto 0);
        rc1            : out   std_logic_vector(ctl_width  * bank_group - 1 downto 0);
        rd1            : inout std_logic_vector(data_width * bank_group - 1 downto 0);
        ra2            : out   std_logic_vector(addr_width * bank_group - 1 downto 0);
        rc2            : out   std_logic_vector(ctl_width  * bank_group - 1 downto 0);
        rd2            : inout std_logic_vector(data_width * bank_group - 1 downto 0);
        ra3            : out   std_logic_vector(addr_width * bank_group - 1 downto 0);
        rc3            : out   std_logic_vector(ctl_width  * bank_group - 1 downto 0);
        rd3            : inout std_logic_vector(data_width * bank_group - 1 downto 0);
        ra4            : out   std_logic_vector(addr_width * bank_group - 1 downto 0);
        rc4            : out   std_logic_vector(ctl_width  * bank_group - 1 downto 0);
        rd4            : inout std_logic_vector(data_width * bank_group - 1 downto 0);
        ra5            : out   std_logic_vector(addr_width * bank_group - 1 downto 0);
        rc5            : out   std_logic_vector(ctl_width  * bank_group - 1 downto 0);
        rd5            : inout std_logic_vector(data_width * bank_group - 1 downto 0);
        ra6            : out   std_logic_vector(addr_width * bank_group - 1 downto 0);
        rc6            : out   std_logic_vector(ctl_width  * bank_group - 1 downto 0);
        rd6            : inout std_logic_vector(data_width * bank_group - 1 downto 0);
        ra7            : out   std_logic_vector(addr_width * bank_group - 1 downto 0);
        rc7            : out   std_logic_vector(ctl_width  * bank_group - 1 downto 0);
        rd7            : inout std_logic_vector(data_width * bank_group - 1 downto 0);
        glob_lbo_l     : out   std_logic;
        glob_flothru_l : out   std_logic);
end zbt64_main;

architecture mixed of zbt64_main is

    function log2(
        x : in natural)
        return natural is
    begin
        for i in 0 to 31 loop
            if x <= 2**i then
                return i;
            end if;
        end loop;
        
        assert false
            report "zbt.vhd: log2() invalid argument"
            severity failure;
        return 0;
    end;

    function min(
        a, b : in integer)
        return integer is
    begin
        if a < b then
            return a;
        else
            return b;
        end if;
    end;

    --
    -- We support up to 8 logical banks of SSRAM.
    --
    constant max_bank         : natural := 8;

    --
    -- We have a 2MB window in local bus space where the SSRAM appears.
    -- This window is augmented with a page register.
    --
    constant page_order       : natural := 18;
    constant page_size        : natural := 2**page_order; -- in quadwords

    --
    -- The number of address bits and maximum size (in longwords) of a
    -- logical SSRAM bank.
    --
    constant bank_order       : natural := addr_width;
    constant bank_size        : natural := 2**bank_order; -- in SSRAM words (quadwords)

    --
    -- We terminate bursts at a page boundary or the smallest possible bank
    -- boundary, whichever occurs more frequently.
    --
    constant bterm_bound      : natural := min(128*1024, page_size); -- in quadwords
    constant bterm_order      : natural := log2(bterm_bound);
    constant bterm_bound_vec  : std_logic_vector(bterm_order - 1 downto 0) :=
        EXT(CONV_STD_LOGIC_VECTOR(bterm_bound - 1, bterm_order + 1), bterm_order);

    signal rst          : std_logic;
    signal clk          : std_logic;
    signal locked       : std_logic_vector(31 downto 0);

    signal lads_i       : std_logic;
    signal lblast_i     : std_logic;
    signal qlads        : std_logic;
    signal lwrite_i     : std_logic;
    signal lready_o     : std_logic;
    signal lready_oe    : std_logic;
    signal lbterm_o     : std_logic;
    signal lbterm_oe    : std_logic;
    signal lbe_i        : std_logic_vector(7 downto 0);

    signal la_q         : std_logic_vector(23 downto 3);

    signal id           : std_logic_vector(63 downto 0);

    signal sram_rd_ready : std_logic;
    signal sram_wr_ready : std_logic;
    signal regs_rd_ready : std_logic;
    signal regs_wr_ready : std_logic;

    signal ds_xfer      : std_logic;
    signal ds_decode    : std_logic;
    signal ds_write     : std_logic;
    signal ds_ready     : std_logic;
    signal ds_stop      : std_logic;
    
    signal logic0       : std_logic;
    signal logic1       : std_logic;

    signal oe_page_reg  : std_logic;    
    signal oe_pipeline_reg : std_logic;    
    signal oe_size_reg  : std_logic;    
    signal oe_info_reg  : std_logic;    
    signal oe_status_reg : std_logic;    
    signal page_reg     : std_logic_vector(7 downto 0);
    signal pipeline_reg : std_logic_vector(0 downto 0);
    signal size_reg     : std_logic_vector(1 downto 0);
    signal info_reg     : std_logic_vector(31 downto 0);
    signal status_reg   : std_logic_vector(31 downto 0);

    signal page_addr    : std_logic_vector(page_order - 1 downto 0);
    signal inc_page_addr : std_logic;
    signal ssram_addr   : std_logic_vector(page_order + 8 - 1 downto 0);

    signal decode_addr  : std_logic_vector(page_order + 8 - 1 downto 0);
    signal banksel_addr : std_logic_vector(2 downto 0);
    signal banksel      : std_logic_vector(max_bank - 1 downto 0);

    signal read_enable  : std_logic;

    type zbt_ra_type is array(0 to max_bank - 1) of
        std_logic_vector(addr_width * bank_group - 1 downto 0);
    type zbt_rc_type is array(0 to max_bank - 1) of
        std_logic_vector(ctl_width * bank_group - 1 downto 0);
    type zbt_rd_type is array(0 to max_bank - 1) of
        std_logic_vector(data_width * bank_group - 1 downto 0);
    type zbt_q_type is array(0 to max_bank - 1) of
        std_logic_vector(63 downto 0);
    signal zbt_read     : std_logic;
    signal zbt_readv    : std_logic_vector(max_bank - 1 downto 0);
    signal zbt_write    : std_logic;
    signal zbt_writev   : std_logic_vector(max_bank - 1 downto 0);
    signal zbt_evalid   : std_logic_vector(max_bank - 1 downto 0);
    signal zbt_q        : zbt_q_type;
    signal oe_zbt_q     : std_logic_vector(max_bank - 1 downto 0);
    signal zbt_t_ra     : zbt_ra_type;
    signal zbt_ra_o     : zbt_ra_type;
    signal zbt_t_rc     : zbt_rc_type;
    signal zbt_rc_o     : zbt_rc_type;
    signal zbt_t_rd     : zbt_rd_type;
    signal zbt_rd_o     : zbt_rd_type;
    signal zbt_rd_i     : zbt_rd_type;

    component clocks
        generic(
            num_clock:    in    natural);
        port(
            lreset_l:     in    std_logic;
            rst:          out   std_logic;
            lclk:         in    std_logic;
            mclk:         in    std_logic;
            clk:          out   std_logic;
            ramclko:      out   std_logic_vector(num_clock - 1 downto 0);
            ramclki:      in    std_logic_vector(num_clock - 1 downto 0);
            locked:       out   std_logic_vector(31 downto 0));
    end component;
        
    component zbt_dpins is
        generic(
            data_width:  in         natural);
        port(
            o:           in         std_logic_vector(data_width - 1 downto 0);
            i:           out        std_logic_vector(data_width - 1 downto 0);
            t:           in         std_logic_vector(data_width - 1 downto 0);
            io:          inout      std_logic_vector(data_width - 1 downto 0));
    end component;

    component zbt_port
        generic(
            addr_width : natural;
            data_width : natural;
            ctl_width  : natural;
            bank_group : natural);
        port(
            clk      : in     std_logic;
            rst      : in     std_logic;
            sr       : in     std_logic;
            pipeline : in     std_logic;
            read     : in     std_logic;
            write    : in     std_logic;
            a        : in     std_logic_vector(addr_width - 1 downto 0);
            d        : in     std_logic_vector(63 downto 0);
            be       : in     std_logic_vector(7 downto 0);
            q        : out    std_logic_vector(63 downto 0);
            evalid   : out    std_logic;
            t_ra     : out    std_logic_vector(addr_width * bank_group - 1 downto 0);
            ra_o     : out    std_logic_vector(addr_width * bank_group - 1 downto 0);
            t_rc     : out    std_logic_vector(ctl_width  * bank_group - 1 downto 0);
            rc_o     : out    std_logic_vector(ctl_width  * bank_group - 1 downto 0);
            t_rd     : out    std_logic_vector(data_width * bank_group - 1 downto 0);
            rd_o     : out    std_logic_vector(data_width * bank_group - 1 downto 0);
            rd_i     : in     std_logic_vector(data_width * bank_group - 1 downto 0));
    end component;

begin

    assert addr_width >= 17 and addr_width <= 20
        report "*** zbt_main: generic 'addr_width' is invalid"
        severity failure;
    
    --
    --	define constant values
    --
    logic0 <= '0';
    logic1 <= '1';

    -- =============================================================================
    -- Local bus interface
    -- =============================================================================

    --
    --	Convert the inputs to active high.
    --	
    lads_i   <= not lads_l;
    lblast_i <= not lblast_l;
    lwrite_i <= lwrite;
    lbe_i    <= not lbe_l;

    --
    -- Generate the local bus output data
    --
    gen_ld_o: process(rst, clk)
    begin
        if rst = '1' then
            ld_o <= (others => '0');
        elsif clk'event and clk = '1' then
            ld_o <= id;
        end if;
    end process;

    --
    -- 'la_q' mirrors 'la' during the burst. To improve timing, rather than using
    -- 'la' direct from the local bus, we latch it at the beginning of the burst and
    -- increment it during the burst.
    --
    -- Note that 'la' cannot be used as the address output to the SSRAMs, as
    -- there are potentially several cycles of latency between data on the local
    -- bus and the same data being read/written from/to a particular SSRAM bank.
    --
    latch_address : process(rst, clk)
    begin
        if rst = '1' then
            la_q <= (others => '0');
        elsif clk'event and clk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
            else
                if ds_xfer = '1' then
                    la_q <= la_q + 1;
                end if;
            end if;
        end if;
    end process;

    --
    -- Generate the 'ds_ready' signal to the direct slave state machine.
    --
    -- This is generated according to the region being addressed by the current burst,
    -- and also whether a read or write is being performed:
    --
    --   o For a read from the SSRAM region, this is generated from the 'zbt_evalid'
    --     vector with 1 cycle of delay.
    --   o For a read from the registers region, this is generated from 'ds_decode'
    --     with 1 cycle of delay.
    --   o For a write to the SSRAM region, this is simply 'ds_decode'.
    --   o For a write to the register region, this is simply 'ds_decode'.
    --
    -- Since 'id' is registered before it is output on the local bus, reads require
    --  1 cycle of delay before 'ds_ready' is asserted, which is why 'regs_rd_ready'
    -- and 'sram_rd_ready' are implemented as registers.
    --
    gen_rd_ready: process(rst, clk)
    begin
        if rst = '1' then
            regs_rd_ready <= '0';
            sram_rd_ready <= '0';
        elsif clk'event and clk = '1' then
            regs_rd_ready <= ds_decode and not ds_write and not la_q(21);
            sram_rd_ready <= read_enable and OR_reduce(zbt_evalid and banksel);
        end if;
    end process;
    regs_wr_ready <= ds_decode and ds_write and not la_q(21);
    sram_wr_ready <= ds_decode and ds_write and la_q(21);
    ds_ready <= sram_rd_ready or sram_wr_ready or regs_rd_ready or regs_wr_ready;
            
    --
    -- Generate the 'ds_stop' signal to the direct slave state machine.
    --
    -- Terminates a burst if the any of the following conditions are met:
    --
    --   o burst is addressing the register region (first case)
    --   o burst is addressing the SSRAM region AND a page or bank boundary
    --     has been reached (last two cases)
    --
    ds_stop <= '1' when ds_decode = '1' and la_q(21) = '0' else
               '1' when ds_decode = '1' and la_q(21) = '1' and la_q(bterm_order + 2 downto 3) = bterm_bound_vec else
               '1' when ds_xfer = '1'   and la_q(21) = '1' and la_q(bterm_order + 2 downto 3) = (bterm_bound_vec - 1) else
               '0';

    --
    -- Generate a qualified address strobe, ie. 'lads_l' qualified by address
    -- decoding for the FPGA space.
    --
    qlads <= lads_i and not la_i(23) and not fholda;

    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed in direct
    -- slave mode, since other local bus masters on the card might also drive it.
    --
    lbterm_l <= not lbterm_o when lbterm_oe = '1' else 'Z';

    --
    -- 'lready_l' should only be driven when the FPGA is addressed in direct
    -- slave mode, since other local bus masters on the card might also drive it.
    --
    lready_l <= not lready_o when lready_oe = '1' else 'Z';

    --
    -- Instantiate the state machine for responding to direct slave transfers.
    --
    plxdssm0: plxdssm
        port map(
            clk        => clk,
            rst        => rst,
            sr         => logic0,
            qlads      => qlads,
            lblast     => lblast_i,
            lwrite     => lwrite_i,
            ld_oe      => oe_ld,
            lready     => lready_o,
            lready_oe  => lready_oe,
            lbterm     => lbterm_o,
            lbterm_oe  => lbterm_oe,
            transfer   => ds_xfer,
            decode     => ds_decode,
            write      => ds_write,
            ready      => ds_ready,
            stop       => ds_stop);

    -- =============================================================================
    -- FPGA Registers
    -- =============================================================================

    --
    -- Generate output enable signals for registers
    --
    gen_oe_regs : process(rst, clk)
    begin
        if rst = '1' then
            oe_zbt_q <= (others => '0');
            oe_page_reg <= '0';
            oe_size_reg <= '0';
            oe_pipeline_reg <= '0';
            oe_info_reg <= '0';
            oe_status_reg <= '0';
        elsif clk'event and clk = '1' then
            if ds_decode = '1' then
                for i in 0 to max_bank - 1 loop
                    if la_q(22 downto 21) = "01" and ds_write = '0' and banksel(i) = '1' then
                        oe_zbt_q(i) <= '1';
                    end if;
                end loop;

                if la_q(22 downto 21) = "00" and ds_write = '0' and la_q(4 downto 3) = "00" then
		    oe_page_reg <= '1';
		    oe_pipeline_reg <= '1';
                end if;

                if la_q(22 downto 21) = "00" and ds_write = '0' and la_q(4 downto 3) = "01" then
		    oe_size_reg <= '1';
                end if;

                if la_q(22 downto 21) = "00" and ds_write = '0' and la_q(4 downto 3) = "10" then
		    oe_info_reg <= '1';
		    oe_status_reg <= '1';
                end if;
            else
                if ds_xfer = '1' and (lblast_i = '1' or lbterm_o = '1') then
                    oe_zbt_q <= (others => '0');
                    oe_page_reg <= '0';
                    oe_pipeline_reg <= '0';
                    oe_size_reg <= '0';
                    oe_info_reg <= '0';
                    oe_status_reg <= '0';
                end if;
            end if;
        end if;
    end process;

    --
    -- Implement the read only 'info_reg' register
    --
    info_reg <= CONV_STD_LOGIC_VECTOR(num_bank, 8) & CONV_STD_LOGIC_VECTOR(bank_size, 24);

    --
    -- Implement the read only 'status_reg' register
    --
    gen_status_reg: process(rst, clk)
    begin
        if rst = '1' then
            status_reg <= (others => '0');
        elsif clk'event and clk = '1' then
            status_reg <= locked;
        end if;
    end process;

    --
    -- Implement the 'page', 'pipeline' and 'size' registers
    --
    registers : process(rst, clk)
    begin
        if rst = '1' then
            page_reg <= (others => '0');
            pipeline_reg <= (others => '0');
            size_reg <= (others => '1');
        elsif clk'event and clk = '1' then
            if ds_xfer = '1' and ds_write = '1' then
                if la_q(22 downto 21) = "00" and la_q(4 downto 3) = "00" then
                    if lbe_i(0) = '1' then
                        page_reg <= ld_i(7 downto 0);
                    end if;

                    if lbe_i(4) = '1' then
                        pipeline_reg <= ld_i(32 downto 32);
                    end if;
                end if;

                if la_q(22 downto 21) = "00" and la_q(4 downto 3) = "01" then
                    if lbe_i(0) = '1' then
                        size_reg <= ld_i(1 downto 0);
                    end if;
                end if;
            end if;
        end if;
    end process;

    --
    -- Drive the internal data bus 'id' on reads.
    --
    gen_id : for i in 0 to max_bank - 1 generate
        id <= zbt_q(i) when oe_zbt_q(i) = '1' else (others => 'Z');
    end generate;
    id(31 downto 0)  <= EXT(page_reg, 32)     when oe_page_reg     = '1' else (others => 'Z');
    id(63 downto 32) <= EXT(pipeline_reg, 32) when oe_pipeline_reg = '1' else (others => 'Z');
    id(31 downto 0)  <= EXT(size_reg, 32)     when oe_size_reg     = '1' else (others => 'Z');
    id(31 downto 0)  <= info_reg              when oe_info_reg     = '1' else (others => 'Z');
    id(63 downto 32) <= status_reg            when oe_status_reg   = '1' else (others => 'Z');

    -- =============================================================================
    -- SSRAM interface
    -- =============================================================================

    --
    -- The 'page_addr' value must be incremented with every time:
    --
    --   o A word is read from the SSRAM
    --   o A word is written to the SSRAM
    --
    inc_page_addr <= zbt_read or zbt_write;

    --
    -- Generate 'page_addr', which is the SSRAM address within the current page.
    --
    --   o At the beginning of the burst, the local bus address is latched into 'page_addr'
    --   o Each time a word is read from or written to the SSRAM ('inc_page_addr'),
    --     'page_addr' is incremented
    --
    gen_page_addr : process(rst, clk)
    begin
        if rst = '1' then
            page_addr <= (others => '0');
        elsif clk'event and clk = '1' then
            if ds_decode = '1' and la_q(22 downto 21) = "01" then
                page_addr <= la_q(page_order + 2 downto 3);
            else
                if inc_page_addr = '1' then
                    page_addr <= page_addr + 1;
                end if;
            end if;
        end if;
    end process;

    --
    -- Generate the address that is output to the SSRAMs, by augmenting 'page_addr'
    -- with the value of the PAGE register.
    --
    -- Note that we ensure that we never burst across a page or bank boundary, by
    -- asserting 'lbterm_l' at appropriate times (see generation of 'lbterm_l' above).
    --
    ssram_addr <= page_reg & page_addr;

    --
    -- Generate 'decode_addr', from the page register and the local bus address.
    -- This value is used during address decoding to identify the bank being accessed.
    --
    decode_addr <= page_reg & la_i(page_order + 2 downto 3);

    --
    -- Pick out the bits of 'decode_addr' that identify the bank being accessed. Which
    -- particular bits are used depends upon the value of the SIZE register.
    --
    gen_banksel_addr: process(
        decode_addr,
        size_reg)
    begin
        case size_reg is
            when "00" =>
                -- SSRAMs are 128k by n
                banksel_addr <= decode_addr(19 downto 17);
            when "01" =>
                -- SSRAMs are 256k by n
                banksel_addr <= decode_addr(20 downto 18); 
            when "10" =>
                -- SSRAMs are 512k by n
                banksel_addr <= decode_addr(21 downto 19);
            when "11" =>
                -- SSRAMs are 1M by n
                banksel_addr <= decode_addr(22 downto 20);
            when others =>
                banksel_addr <= (others => '-');
        end case;
    end process;
         
    --
    -- Generate the one-hot 'banksel' vector, from 'banksel_addr', which
    -- indicates whether or not a particular bank is being accessed.
    --
    gen_banksel : process(rst, clk)
    begin
        if rst = '1' then
            banksel <= (others => '0');
        elsif clk'event and clk = '1' then
            if lads_i = '1' then
                for i in 0 to max_bank - 1 loop
                    if banksel_addr = i then
                        banksel(i) <= '1';
                    else
                        banksel(i) <= '0';
                    end if;
                end loop;
            end if;
        end if;
    end process;

    --
    -- Generate the 'reading' signal, which is asserted on reads from cycle
    -- following the 'lads_l' pulse until the end of the current burst.
    --
    -- This signal is used to tell a selected zbt_port module to begin reading.
    --
    gen_reading : process(rst, clk)
    begin
        if rst = '1' then
            read_enable <= '0';
        elsif clk'event and clk = '1' then
            if ds_xfer = '1' and (lblast_i = '1' or lbterm_o = '1') then
                read_enable <= '0';
            else
                if ds_decode = '1' and ds_write = '0' and la_q(22 downto 21) = "01" then
                    read_enable <= '1';
                end if;
            end if;
        end if;
    end process;

    --
    -- Generate the one-hot 'zbt_writev' and 'zbt_readv' vectors to
    -- the zbt_port modules.
    --
    zbt_write <= ds_xfer and ds_write;
    gen_zbt_writev : for i in 0 to max_bank - 1 generate
        zbt_writev(i) <= '1' when ds_xfer = '1' and ds_write = '1' and banksel(i) = '1' and la_q(22 downto 21) = "01" else
                         '0';
    end generate;

    gen_zbt_read : process(rst, clk)
    begin
        if rst = '1' then
            zbt_readv <= (others => '0');
            zbt_read  <= '0';
        elsif clk'event and clk = '1' then
            for i in 0 to max_bank - 1 loop
                zbt_readv(i) <= read_enable and banksel(i);
            end loop;

            zbt_read <= read_enable;
        end if;
    end process;

    --
    -- Instantiate the ZBT ports (bank interfaces)
    --
    generate_zbt_ports: for i in 0 to num_bank - 1 generate
        zbt_port0: zbt_port
            generic map(
                addr_width => addr_width,
                data_width => data_width,
                ctl_width  => ctl_width,
                bank_group => bank_group)
            port map(
                clk      => clk,
                rst      => rst,
                sr       => logic0,
                pipeline => pipeline_reg(0),
                read     => zbt_readv(i),
                write    => zbt_writev(i),
                a        => ssram_addr(bank_order - 1 downto 0),
                d        => ld_i,
                be       => lbe_i,
                q        => zbt_q(i),
                evalid   => zbt_evalid(i),
                t_ra     => zbt_t_ra(i),
                ra_o     => zbt_ra_o(i),
                t_rc     => zbt_t_rc(i),
                rc_o     => zbt_rc_o(i),
                t_rd     => zbt_t_rd(i),
                rd_o     => zbt_rd_o(i),
                rd_i     => zbt_rd_i(i));
    end generate;

    --
    -- For banks not present, generate dummy signals.
    --
    generate_zbt_dummy : for i in num_bank to max_bank - 1 generate
        zbt_t_ra(i)    <= (others => '1');
        zbt_ra_o(i)    <= (others => '0');
        zbt_t_rc(i)    <= (others => '1');
        zbt_rc_o(i)    <= (others => '1');
        zbt_t_rd(i)    <= (others => '1');
        zbt_rd_o(i)    <= (others => '0');
	zbt_evalid(i)  <= read_enable;
        zbt_q(i)       <= (others => '0');
    end generate;

    --
    -- Instantiate the 'zbt_dpins' modules. These are used to give a
    -- drive strength of 16mA as recommended in a reference ZBT interface
    -- design from Xilinx.
    --
    iobufs0: zbt_dpins
        generic map(data_width => data_width * bank_group)
        port map(i => zbt_rd_i(0), o => zbt_rd_o(0), t => zbt_t_rd(0), io => rd0);
    iobufs1: zbt_dpins
        generic map(data_width => data_width * bank_group)
        port map(i => zbt_rd_i(1), o => zbt_rd_o(1), t => zbt_t_rd(1), io => rd1);
    iobufs2: zbt_dpins
        generic map(data_width => data_width * bank_group)
        port map(i => zbt_rd_i(2), o => zbt_rd_o(2), t => zbt_t_rd(2), io => rd2);
    iobufs3: zbt_dpins
        generic map(data_width => data_width * bank_group)
        port map(i => zbt_rd_i(3), o => zbt_rd_o(3), t => zbt_t_rd(3), io => rd3);
    iobufs4: zbt_dpins
        generic map(data_width => data_width * bank_group)
        port map(i => zbt_rd_i(4), o => zbt_rd_o(4), t => zbt_t_rd(4), io => rd4);
    iobufs5: zbt_dpins
        generic map(data_width => data_width * bank_group)
        port map(i => zbt_rd_i(5), o => zbt_rd_o(5), t => zbt_t_rd(5), io => rd5);
    iobufs6: zbt_dpins
        generic map(data_width => data_width * bank_group)
        port map(i => zbt_rd_i(6), o => zbt_rd_o(6), t => zbt_t_rd(6), io => rd6);
    iobufs7: zbt_dpins
        generic map(data_width => data_width * bank_group)
        port map(i => zbt_rd_i(7), o => zbt_rd_o(7), t => zbt_t_rd(7), io => rd7);
           
    --
    -- Output the address and control busses to the SSRAMs
    --
    generate_ra : for i in 0 to addr_width * bank_group - 1 generate
        ra0(i) <= zbt_ra_o(0)(i) when zbt_t_ra(0)(i) = '0' else 'Z';
        ra1(i) <= zbt_ra_o(1)(i) when zbt_t_ra(1)(i) = '0' else 'Z';
        ra2(i) <= zbt_ra_o(2)(i) when zbt_t_ra(2)(i) = '0' else 'Z';
        ra3(i) <= zbt_ra_o(3)(i) when zbt_t_ra(3)(i) = '0' else 'Z';
        ra4(i) <= zbt_ra_o(4)(i) when zbt_t_ra(4)(i) = '0' else 'Z';
        ra5(i) <= zbt_ra_o(5)(i) when zbt_t_ra(5)(i) = '0' else 'Z';
        ra6(i) <= zbt_ra_o(6)(i) when zbt_t_ra(6)(i) = '0' else 'Z';
        ra7(i) <= zbt_ra_o(7)(i) when zbt_t_ra(7)(i) = '0' else 'Z';
    end generate;

    generate_rc : for i in 0 to ctl_width * bank_group - 1 generate
        rc0(i) <= zbt_rc_o(0)(i) when zbt_t_rc(0)(i) = '0' else 'Z';
        rc1(i) <= zbt_rc_o(1)(i) when zbt_t_rc(1)(i) = '0' else 'Z';
        rc2(i) <= zbt_rc_o(2)(i) when zbt_t_rc(2)(i) = '0' else 'Z';
        rc3(i) <= zbt_rc_o(3)(i) when zbt_t_rc(3)(i) = '0' else 'Z';
        rc4(i) <= zbt_rc_o(4)(i) when zbt_t_rc(4)(i) = '0' else 'Z';
        rc5(i) <= zbt_rc_o(5)(i) when zbt_t_rc(5)(i) = '0' else 'Z';
        rc6(i) <= zbt_rc_o(6)(i) when zbt_t_rc(6)(i) = '0' else 'Z';
        rc7(i) <= zbt_rc_o(7)(i) when zbt_t_rc(7)(i) = '0' else 'Z';
    end generate;

    --
    -- Output global operating modes to SSRAMs
    --
    glob_lbo_l     <= '0';
    glob_flothru_l <= pipeline_reg(0);

    --
    -- Instantiate the module that generates the SSRAM clocks
    --		
    clocks0: clocks
        generic map(
            num_clock => num_clock)
        port map(
            lreset_l  => lreset_l,
            rst       => rst,
            lclk      => lclk,
            mclk      => mclk,
            clk       => clk,
            ramclko   => ramclko,
            ramclki   => ramclki,
            locked    => locked);
            
end mixed;
